#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""IPXACT Parser."""

from collections.abc import Iterator
from functools import lru_cache
from pathlib import Path

import ucdp as u
import xmlschema
from defusedxml import ElementTree
from ucdp.docutil import doc_from_type
from ucdp.humannum import Hex
from ucdp.orientation import IN, INOUT, OUT
from ucdp.signal import Port
from ucdp.typescalar import BitType, UintType
from ucdp_addr.addrspace import Addrspace
from ucdp_ipxact import Parser

# generated with: xsdata spirit_1685_2009/xml_schema/component.xsd -ss single-package -p schema_dataclass_pkg
from ucdp_ipxact.spirit_1685_2009.xml_schema_pkg import Component
from xsdata.formats.dataclass.parsers import XmlParser

DIRECTION_MAP = {
    "in": IN,
    "out": OUT,
    "inout": INOUT,
}

ACCESS_MAP = {
    "read-only": "RO",
    "write-only": "WO",
    "read-write": "RW",
    "writeOnce": "WO",
    "read-writeOnce": "RW",
}


class GenericIpxactEnumType(u.AEnumType):
    """Generic Enum Type for Ipxact Enumerated Values."""

    keytype: u.UintType = u.UintType(8)
    iter_enumerated_values: tuple[tuple[int, str, str], ...] = u.Field(repr=False)

    def _build(self) -> None:
        for (
            idx,
            name,
            title,
        ) in self.iter_enumerated_values:
            self._add(idx, name, title=title)


@lru_cache
def _is_compatible(filepath: Path) -> bool:
    try:
        etree = ElementTree.parse(filepath)
    except Exception:
        return False
    root = etree.getroot().tag
    return "XMLSchema/SPIRIT/1685-2009}" in root


class Spirit2009Parser(Parser):
    """General IPXACT Parser."""

    @staticmethod
    def is_compatible(filepath: Path) -> bool:
        """Check if Parsable."""
        return _is_compatible(filepath)

    @staticmethod
    def validate(filepath: Path) -> bool:
        """Validate."""
        # local copy of schema: xsdata download http://www.spiritconsortium.org/XMLSchema/SPIRIT/1685-2009/index.xsd
        ipxact_schema = xmlschema.XMLSchema(Path(__file__).parent / "xml_schema" / "index.xsd")
        return ipxact_schema.is_valid(filepath)

    def _read(self, filepath: Path):
        """Map xml on Component Dataclass."""
        xml_parser = XmlParser()
        return xml_parser.parse(filepath, Component)

    def _get_vendor(self, data) -> str:
        return data.vendor

    def _get_name(self, data) -> str:
        return data.name

    def _get_library(self, data) -> str:
        return data.library

    def _get_version(self, data) -> str:
        return data.version

    def _get_ports(self, data) -> Iterator[u.Port]:
        for businterface in self._iter_bus_interfaces(data):
            for port_map in self._iter_port_maps(businterface):
                pys_port = self._find_phys_port(data, port_map.physical_port.name)
                port_direction = DIRECTION_MAP[pys_port.wire.direction.value]
                port_type = self._get_port_type(pys_port.wire.vector)
                port_comment = self._get_log_port_name(port_map)
                port_descr = None
                if pys_port.description:
                    port_descr = pys_port.description.value
                port_doc = doc_from_type(port_type, comment=port_comment, descr=port_descr)
                yield Port(port_type, pys_port.name, direction=port_direction, doc=port_doc)

    def _iter_bus_interfaces(self, comp):
        yield from comp.bus_interfaces.bus_interface

    def _iter_port_maps(self, bus_interface):
        yield from bus_interface.port_maps.port_map

    def _find_phys_port(self, comp, phys_port_name):
        for port in comp.model.ports.port:
            if port.name == phys_port_name:
                return port
        raise ValueError(f"No Physical Port {phys_port_name} found.")

    def _get_port_type(self, port_vector):
        port_type = BitType()
        if port_vector:
            left = int(port_vector.left.value)
            right = int(port_vector.right.value)
            port_type = UintType(width=left - right + 1, right=right)
        return port_type

    def _get_log_port_name(self, port_map) -> str:
        name = port_map.logical_port.name
        vector = port_map.logical_port.vector
        if vector:
            left = int(vector.left.value)
            right = int(vector.right.value)
            name = name + f"[{left}:{right}]"
        return name

    def _get_addrspaces(self, data) -> Iterator[Addrspace]:
        for memory_map in self._iter_memory_maps(data):
            for addr_block in self._iter_addr_blocks(memory_map):
                addrspace = self._get_addrspace(addr_block)
                for register in self._iter_registers(addr_block):
                    word = self._add_word(addrspace, register)
                    word_default = self._get_word_default(register)
                    for field in self._iter_fields(register):
                        self._add_field(word, field, word_default)
                yield addrspace

    def _get_word_default(self, register):
        mask = register.reset.mask
        word_default = Hex(register.reset.value.value)
        if mask:
            word_default = word_default & Hex(mask.value)
        return word_default

    def _get_field_default(self, field, word_default):
        slice_right = field.bit_offset
        slice_left = slice_right + field.bit_width.value - 1
        slice = u.slices.Slice(left=slice_left, right=slice_right)
        return slice.extract(word_default)

    def _add_field(self, word, field, word_default):
        word.add_field(
            field.name.lower(),
            self._get_field_type(field, self._get_field_default(field, word_default)),
            ACCESS_MAP[field.access.value.value],
            offset=field.bit_offset,
            descr=field.description.value,
            is_volatile=self._get_volatile(field),
        )

    def _iter_fields(self, register):
        yield from register.field_value

    def _add_word(self, addrspace, register):
        return addrspace.add_word(
            name=register.name.lower(),
            offset=Hex(register.address_offset),
            descr=register.description.value,
            title=register.display_name.value,
            is_volatile=self._get_volatile(register),
        )

    def _get_field_enum_type(self, field):
        for val in field.enumerated_values.enumerated_value:
            yield (Hex(val.value), val.name, val.display_name.value)

    def _get_field_type(self, field, default):
        field_type = u.UintType(field.bit_width.value, default=default)
        if field.enumerated_values:
            field_type = GenericIpxactEnumType(
                keytype=UintType(field.bit_width.value, default=default),
                iter_enumerated_values=tuple(self._get_field_enum_type(field)),
            )
        return field_type

    def _get_volatile(self, cls) -> bool:
        volatile = False
        if cls.volatile:
            volatile = cls.volatile.value
        return volatile

    def _iter_registers(self, addr_block):
        yield from addr_block.register

    def _get_addrspace(self, addr_block) -> Addrspace:
        addrspace_name = addr_block.name
        addrspace_baseaddr = Hex(addr_block.base_address.value)
        addrspace_size = Hex(addr_block.range.value)
        addrspace_width = addr_block.width.value
        return Addrspace(
            name=addrspace_name,
            baseaddr=addrspace_baseaddr,
            width=addrspace_width,
            size=addrspace_size,
        )

    def _iter_memory_maps(self, comp):
        yield from comp.memory_maps.memory_map

    def _iter_addr_blocks(self, memory_map):
        yield from memory_map.address_block



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""IPXACT Parser."""

from abc import abstractmethod
from collections.abc import Iterator
from pathlib import Path

import ucdp as u
from ucdp_addr.addrspace import Addrspace

from .ipxact import UcdpIpxact


class Parser(u.Object):
    """General IPXACT Parser."""

    prio: int = 0

    @staticmethod
    def is_compatible(filepath: Path) -> bool:
        """Check if Parsable."""
        raise NotImplementedError

    @staticmethod
    def validate(filepath: Path) -> bool:
        """Validate."""
        raise NotImplementedError

    def parse(self, filepath: Path) -> UcdpIpxact:
        """Parse."""
        data = self._read(filepath)
        return UcdpIpxact(
            vendor=self._get_vendor(data),
            version=self._get_version(data),
            libname=self._get_library(data),
            name=self._get_name(data),
            ports=tuple(self._get_ports(data)),
            addrspaces=tuple(self._get_addrspaces(data)),
        )

    @abstractmethod
    def _read(self, filepath: Path):
        raise NotImplementedError

    @abstractmethod
    def _get_vendor(self, data) -> str:
        raise NotImplementedError

    @abstractmethod
    def _get_name(self, data) -> str:
        raise NotImplementedError

    @abstractmethod
    def _get_library(self, data) -> str:
        raise NotImplementedError

    @abstractmethod
    def _get_version(self, data) -> str:
        raise NotImplementedError

    @abstractmethod
    def _get_ports(self, data) -> Iterator[u.Port]:
        raise NotImplementedError

    @abstractmethod
    def _get_addrspaces(self, data) -> Iterator[Addrspace]:
        raise NotImplementedError



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Parser Manager."""

import re

from pytest import raises
from ucdp_ipxact.parsermanager import ParserManager


def test_basics():
    """Basic Testing."""
    parsermanager = ParserManager()
    assert len(parsermanager.parsers) == 0


def test_create():
    """Check Parser Manager Initialization."""
    parsermanager = ParserManager.create()
    assert len(parsermanager.parsers) == 1


def test_compatible():
    """Check Parser Manager is_compatible."""
    parsermanager = ParserManager.create()
    assert parsermanager.is_compatible("tests/testdata/example-2009.xml") is True
    assert parsermanager.is_compatible("tests/testdata/empty.xml") is False


def test_validate():
    """Check Parser Manager validate."""
    parsermanager = ParserManager.create()
    assert parsermanager.validate("tests/testdata/example-2009.xml") is True

    msg = "Unknown schema in"
    with raises(ValueError, match=re.escape(msg)):
        parsermanager.validate("tests/testdata/empty.xml")



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Command Line Interface."""

import sys
from contextlib import contextmanager
from pathlib import Path

import click
import ucdp as u

from ucdp_ipxact import UcdpIpxact, get_parser

ipxact = u.cli.get_group("IPXACT Commands")


@contextmanager
def _load(console, filepath: Path) -> UcdpIpxact:
    try:
        parser = get_parser(filepath)
        parser.validate(filepath)
        ucdp_ipxact = parser.parse(filepath)
    except Exception as exc:
        print(f"ERROR: {exc}", file=sys.stderr)
        print(f"'{filepath!s}' is INVALID.")
        sys.exit(1)
    yield ucdp_ipxact


@ipxact.command()
@click.argument("ipxact", type=click.Path(exists=True, path_type=Path))
@u.cli.pass_ctx
def check(ctx, ipxact: Path):
    """Validate IPXACT and convert to UCDP Format."""
    with _load(ctx.console, ipxact):
        print(f"'{ipxact!s}' is valid.")


@ipxact.command()
@click.argument("ipxact", type=click.Path(exists=True, path_type=Path))
@u.cli.pass_ctx
def overview(ctx, ipxact: Path):
    """Load IPXACT and Show Overview."""
    with _load(ctx.console, ipxact) as ucdp_result:
        print(ucdp_result.get_overview())



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""IPXACT Version Manager."""

import importlib
import logging
from importlib.metadata import entry_points
from pathlib import Path

import ucdp as u

from .ipxact import UcdpIpxact
from .parser import Parser

LOGGER = logging.getLogger(__name__)


class ParserManager(u.Object):
    """Parser Manager collects all available IPXACT Parser."""

    _parsers: list[Parser] = u.PrivateField(default_factory=list)

    @staticmethod
    def create() -> "ParserManager":
        """Create ParserManager with all IPXACT-Parser."""
        parsermanager = ParserManager()
        for entry_point in entry_points(group="ucdp_ipxact.parser"):
            parsercls = _load(entry_point.value)
            if not issubclass(parsercls, Parser):  # pragma: no cover
                LOGGER.warning("Entrypoint %s %s is not a IPXACT-Parser", entry_point.name, entry_point.value)
                continue
            parsermanager.add(parsercls())
        return parsermanager

    def add(self, parser: Parser) -> None:
        """Add Parsers to ParserManager."""
        self._parsers.append(parser)

    @property
    def parsers(self) -> tuple[Parser, ...]:
        """Parsers."""
        return tuple(self._parsers)

    def is_compatible(self, filepath: Path) -> bool:
        """Check if Parsable."""
        filepath = u.improved_resolve(filepath, strict=True, replace_envvars=True)
        for parser in self._parsers:
            if parser.is_compatible(filepath):
                return True
        return False

    def validate(self, filepath: Path) -> bool:
        """Validate."""
        filepath = u.improved_resolve(filepath, strict=True, replace_envvars=True)
        return self.get_parser(filepath).validate(filepath)

    def parse(self, filepath: Path) -> UcdpIpxact:
        """Parse."""
        filepath = u.improved_resolve(filepath, strict=True, replace_envvars=True)
        return self.get_parser(filepath).parse(filepath)

    def get_parser(self, filepath: Path) -> Parser:
        """Returns first campatible Parser."""
        for parser in sorted(self._parsers, key=lambda parser: parser.prio, reverse=True):
            if parser.is_compatible(filepath):
                return parser
        raise ValueError(f"Unknown schema in {filepath}")


def _load(ref):
    modname, cmd_object_name = ref.rsplit(".", 1)
    mod = importlib.import_module(modname)
    return getattr(mod, cmd_object_name)


def validate(filepath: Path) -> bool:
    """Validate IPXACT."""
    parsermanager = ParserManager.create()
    return parsermanager.validate(filepath)


def parse(filepath: Path) -> UcdpIpxact:
    """Parse IPXACT."""
    parsermanager = ParserManager.create()
    return parsermanager.parse(filepath)


def get_parser(filepath: Path) -> Parser:
    """Determine Parser for IPXACT."""
    parsermanager = ParserManager.create()
    return parsermanager.get_parser(filepath)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test IPXACT Importing."""

from pathlib import Path

import ucdp as u
from pytest import raises
from test2ref import assert_refdata
from ucdp_ipxact import UcdpIpxactMod, parse, validate

TESTDATA_PATH = Path(__file__).parent / "testdata"


class Top2009Mod(u.AMod):
    """2009 Version."""

    def _build(self):
        UcdpIpxactMod(self, "u_inst0", filepath=Path("testdata/example-2009.xml"))
        UcdpIpxactMod(self, "u_inst1", filepath=Path("testdata/example-2009.xml"))


def test_missing_file(tmp_path):
    """Missing."""
    with raises(FileNotFoundError):
        validate(tmp_path / "missing.xml")


def test_validate_2009():
    """Validate 2009 Example Xml."""
    assert validate(TESTDATA_PATH / "example-2009.xml") is True


def test_parse_2009(tmp_path):
    """Parse 2009 Example Xml."""
    filepath = TESTDATA_PATH / "example-2009.xml"
    ipxact = parse(filepath)
    (tmp_path / "result.txt").write_text(ipxact.get_overview())
    assert_refdata(test_parse_2009, tmp_path)


def test_mod_2009(tmp_path):
    """2009 Example Module."""
    mod = Top2009Mod()

    inst0 = mod.get_inst("u_inst0")
    (tmp_path / "inst0.txt").write_text(inst0.get_overview())
    assert inst0.name == "u_inst0"
    assert inst0.modname == "VGA"
    assert inst0.libname == "user"
    assert len(tuple(inst0.get_addrspaces())) == 2

    inst1 = mod.get_inst("u_inst1")
    (tmp_path / "inst1.txt").write_text(inst1.get_overview())
    assert inst1.name == "u_inst1"
    assert inst1.modname == "VGA"
    assert inst1.libname == "user"
    assert len(tuple(inst1.get_addrspaces())) == 2

    assert_refdata(test_mod_2009, tmp_path)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Command-Line-Interface."""

from click.testing import CliRunner
from pytest import fixture
from test2ref import assert_refdata
from ucdp.cli import ucdp


@fixture
def runner():
    """Click Runner for Testing."""
    yield CliRunner()


def _run(runner, path, cmd, exit_code=0):
    result = runner.invoke(ucdp, cmd)
    assert result.exit_code == exit_code
    (path / "console.txt").write_text(result.output.rstrip() + "\n")


def test_check(runner, tmp_path):
    """Check Command."""
    cmd = ["ipxact", "check", "tests/testdata/example-2009.xml"]
    _run(runner, tmp_path, cmd)
    assert_refdata(test_check, tmp_path)


def test_check_empty(runner, tmp_path):
    """Check Command on Empty File."""
    cmd = ["ipxact", "check", "tests/testdata/empty.xml"]
    _run(runner, tmp_path, cmd, exit_code=1)
    assert_refdata(test_check_empty, tmp_path)


def test_overview(runner, tmp_path):
    """Overview Command."""
    cmd = ["ipxact", "overview", "tests/testdata/example-2009.xml"]
    _run(runner, tmp_path, cmd)
    assert_refdata(test_overview, tmp_path)


def test_overview_empty(runner, tmp_path):
    """Overview Command on Empty File."""
    cmd = ["ipxact", "overview", "tests/testdata/empty.xml"]
    _run(runner, tmp_path, cmd, exit_code=1)
    assert_refdata(test_overview_empty, tmp_path)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Unified Chip Design Platform - IPXACT."""

from logging import getLogger

import ucdp as u
from tabulate import tabulate
from ucdp_addr import AddrMap, Addrspace

LOGGER = getLogger(__name__)


class UcdpIpxact(u.Object):
    """UCDP Representation of IPXACT."""

    vendor: str
    version: str
    libname: str
    name: str
    ports: tuple[u.Port, ...] = ()
    addrspaces: tuple[Addrspace, ...] = ()
    filelists: u.ModFileLists = ()

    def get_overview(self) -> str:
        """Nice Overview."""
        addrmap = AddrMap.from_addrspaces(self.addrspaces)
        ports = [(port.direction, port.name, port.type_) for port in self.ports]
        data = (
            ("Attribute", "Value"),
            ("vendor", self.vendor),
            ("version", self.version),
            ("libname", self.libname),
            ("name", self.name),
            ("ports", tabulate(ports, headers=("I/O", "Name", "Type"))),
            ("addrmap", addrmap.get_overview()),
        )
        return tabulate(data, tablefmt="fancy_grid") + "\n"



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Unified Chip Design Platform - IPXACT."""

from collections.abc import Iterator
from functools import cached_property, lru_cache
from inspect import getfile
from logging import getLogger
from pathlib import Path

import ucdp as u
from ucdp_addr import Addrspace

from .ipxact import UcdpIpxact
from .parsermanager import ParserManager

LOGGER = getLogger(__name__)


@lru_cache
def _get_parsermanager() -> ParserManager:
    return ParserManager.create()


@lru_cache
def _parse(filepath: Path) -> UcdpIpxact:
    """Parse IPXACT."""
    LOGGER.info("Reading %s", filepath)
    return _get_parsermanager().parse(filepath)


def _load(mod: u.BaseMod, filepath: Path) -> UcdpIpxact:
    basedir = Path(getfile(mod.parent.__class__)).parent if mod.parent else None
    filepath = u.improved_resolve(filepath, basedir=basedir, strict=True, replace_envvars=True)
    return _parse(filepath)


class UcdpIpxactMod(u.ATailoredMod):
    """IPXACT Import Module."""

    filepath: Path

    @cached_property
    def ipxact(self) -> UcdpIpxact:
        """IPXACT."""
        return _load(self, self.filepath)

    def _build(self):
        ipxact = self.ipxact
        for port in ipxact.ports:
            name = port.name
            self.namespace[name] = port
            self.portssignals[name] = port
            self.ports[name] = port

    @property
    def modname(self) -> str:
        """Module Name."""
        return self.ipxact.name

    @property
    def libname(self) -> str:
        """Library Name."""
        return self.ipxact.libname

    def get_addrspaces(self) -> Iterator[Addrspace]:
        """Address Spaces."""
        yield from self.ipxact.addrspaces

    def get_overview(self) -> str:
        """Overview."""
        return self.ipxact.get_overview()
